#
# Defines data that is consumed by the header2whatever hooks/templates
# to modify the generated files
#

import enum
from typing import Dict, List, Tuple, Optional

from pydantic import BaseModel, validator


class Model(BaseModel):
    class Config:
        extra = "forbid"


class ParamData(Model):
    """Various ways to modify parameters"""

    # C++ type emitted
    x_type: Optional[str] = None

    # Default value for parameter
    default: Optional[str] = None

    force_out: bool = False

    ignore: bool = False


class BufferType(str, enum.Enum):
    IN = "in"
    OUT = "out"
    INOUT = "inout"


class BufferData(Model):
    """Describes buffers"""

    # Indicates what type of buffer is required: out/inout buffers require
    # a writeable buffer such as a bytearray, but in only needs a readonly
    # buffer (such as bytes)
    type: BufferType

    # Name of source parameter -- user must pass in something that implements
    # the buffer protocol (eg bytes, bytearray)
    src: str

    # Name of length parameter. An out-only parameter, it will be set to the size
    # of the src buffer, and will be returned so the caller can determine how
    # many bytes were written
    len: str

    # If specified, the minimum size of the passed in buffer
    minsz: Optional[int] = None


class ReturnValuePolicy(enum.Enum):
    """
        see pybind11 documentation for this
    """

    TAKE_OWNERSHIP = "take_ownership"
    COPY = "copy"
    MOVE = "move"
    REFERENCE = "reference"
    REFERENCE_INTERNAL = "reference_internal"
    AUTOMATIC = "automatic"
    AUTOMATIC_REFERENCE = "automatic_reference"


class FunctionData(Model):
    # If True, don't wrap this
    ignore: bool = False

    # Generate this in an `#ifdef`
    ifdef: Optional[str] = None
    # Generate this in an `#ifndef`
    ifndef: Optional[str] = None

    # Use this code instead of the generated code
    cpp_code: Optional[str] = None

    # Docstring for the function
    doc: Optional[str] = None

    # If True, prepends an underscore to the python name
    internal: bool = False

    # Set this to rename the function
    rename: Optional[str] = None

    # Mechanism to override individual parameters
    param_override: Dict[str, ParamData] = {}

    #: If specified, put the function in a sub.pack.age
    subpackage: Optional[str] = None

    no_release_gil: Optional[bool] = None

    buffers: List[BufferData] = []

    overloads: Dict[str, "FunctionData"] = {}

    #: Adds py::keep_alive<x,y> to the function. Overrides automatic
    #: keepalive support, which retains references passed to constructors
    keepalive: Optional[List[Tuple[int, int]]] = None

    return_value_policy: ReturnValuePolicy = ReturnValuePolicy.AUTOMATIC

    @validator("overloads", pre=True)
    def validate_overloads(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = FunctionData()
        return value


FunctionData.update_forward_refs()


class PropAccess(enum.Enum):

    #: Determine read/read-write automatically
    AUTOMATIC = "auto"

    #: Allow python users access to the value, but ensure it can't
    #: change. This is useful for properties that are defined directly
    #: in the class
    READONLY = "readonly"

    #: Allows python users to read/write the value
    READWRITE = "readwrite"


class PropData(Model):
    ignore: bool = False

    #: Set the name of this property to the specified value
    rename: Optional[str]

    #: Python code access to this property
    access: PropAccess = PropAccess.AUTOMATIC


class EnumValue(Model):
    ignore: bool = False
    rename: Optional[str] = None


class EnumData(Model):
    ignore: bool = False
    rename: Optional[str] = None
    value_prefix: Optional[str] = None

    #: If specified, put the enum in a sub.pack.age (ignored for
    #: enums that are part of classes)
    subpackage: Optional[str] = None

    values: Dict[str, EnumValue] = {}


class ClassData(Model):

    ignore: bool = False
    ignored_bases: List[str] = []

    # Specify fully qualified names for the bases
    base_qualnames: Dict[str, str] = {}

    attributes: Dict[str, PropData] = {}
    enums: Dict[str, EnumData] = {}
    methods: Dict[str, FunctionData] = {}

    is_polymorphic: bool = False
    force_no_trampoline: bool = False
    force_no_default_constructor: bool = False

    #: If there are circular dependencies, this will help you resolve them
    #: manually. TODO: make it so we don't need this
    force_depends: List[str] = []

    #: If the object shouldn't be deleted by pybind11, use this. Disables
    #: implicit constructors.
    nodelete: bool = False

    #: Set the python name of the class to this
    rename: Optional[str] = None

    #: If the type was created as a shared_ptr (such as via std::make_shared)
    #: then pybind11 needs to be informed of this.
    #:
    #: https://github.com/pybind/pybind11/issues/1215
    #:
    #: One way you can tell we messed this up is if there's a double-free
    #: error and the stack trace involves a unique_ptr destructor
    shared_ptr: bool = True

    #: If specified, put the class in a sub.pack.age. Ignored
    #: for functions attached to a class.
    subpackage: Optional[str] = None

    #: Extra 'using' directives to insert into the trampoline and the
    #: wrapping scope
    typealias: List[str] = []

    #: Extra constexpr to insert into the trampoline and wrapping scopes
    constants: List[str] = []

    @validator("attributes", pre=True)
    def validate_attributes(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = PropData()
        return value

    @validator("enums", pre=True)
    def validate_enums(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = EnumData()
        return value

    @validator("methods", pre=True)
    def validate_methods(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = FunctionData()
        return value


class HooksDataYaml(Model):
    """
        Format of the file in [tool.robotpy-build.wrappers."PACKAGENAME"]
        generation_data
    """

    strip_prefixes: List[str] = []
    extra_includes: List[str] = []
    # sticks these includes first, for really broken headers
    extra_includes_first: List[str] = []
    inline_code: Optional[str] = None

    attributes: Dict[str, PropData] = {}
    classes: Dict[str, ClassData] = {}
    functions: Dict[str, FunctionData] = {}
    enums: Dict[str, EnumData] = {}

    @validator("attributes", pre=True)
    def validate_attributes(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = PropData()
        return value

    @validator("classes", pre=True)
    def validate_classes(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = ClassData()
        return value

    @validator("enums", pre=True)
    def validate_enums(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = EnumData()
        return value

    @validator("functions", pre=True)
    def validate_functions(cls, value):
        for k, v in value.items():
            if v is None:
                value[k] = FunctionData()
        return value



# Used by robotpybuild entry point

from os.path import abspath, join, dirname
from typing import Dict, List, Optional

_root = abspath(dirname(__file__))


def get_include_dirs() -> Optional[List[str]]:
    return [join(_root, "pybind11", "include"), join(_root, "include")]


def get_library_dirs() -> Optional[List[str]]:
    pass


def get_type_casters(casters: Dict[str, str]) -> None:
    casters.update(
        {
            "std::vector": "pybind11/stl.h",
            "std::deque": "pybind11/stl.h",
            "std::list": "pybind11/stl.h",
            "std::array": "pybind11/stl.h",
            "std::valarray": "pybind11/stl.h",
            "std::set": "pybind11/stl.h",
            "std::map": "pybind11/stl.h",
            "std::unordered_map": "pybind11/stl.h",
            "std::optional": "pybind11/stl.h",
            "std::nullopt_t": "pybind11/stl.h",
            "std::variant": "pybind11/stl.h",
            "std::function": "pybind11/functional.h",
            "std::complex": "pybind11/complex.h",
            "std::chrono::duration": "pybind11/chrono.h",
            "std::chrono::time_point": "pybind11/chrono.h",
        }
    )



def get_install_root(cmd):
    # hack
    install_root = getattr(cmd.distribution, "rpybuild_develop_path", None)
    if not install_root:
        inst_command = cmd.distribution.get_command_obj("install")
        inst_command.ensure_finalized()
        install_root = inst_command.install_platlib

    return install_root


def get_build_temp_path():

    import distutils.dist
    import distutils.command.build

    b = distutils.command.build.build(distutils.dist.Distribution())
    b.finalize_options()
    return b.build_temp



import sys
import yaml

from .hooks_datacfg import (
    ClassData,
    EnumData,
    HooksDataYaml,
    PropData,
    FunctionData,
)

from typing import Optional


_missing = object()


class GeneratorData:
    """
        Used by the hooks to retrieve user-specified generation data, and 
        report to the user that there is data missing
    """

    data: HooksDataYaml

    def __init__(self, data: HooksDataYaml):
        self.data = data

        # report data
        self.functions = {}
        self.classes = {}
        self.enums = {}
        self.attributes = {}

    def get_class_data(self, name: str) -> ClassData:
        data = self.data.classes.get(name, _missing)
        missing = data is _missing
        if missing:
            data = ClassData()

        self.classes[name] = {
            "attributes": {},
            "enums": {},
            "functions": {},
            "missing": missing,
        }
        return data

    def get_cls_enum_data(
        self, name: str, cls_key: str, cls_data: ClassData
    ) -> EnumData:
        if name is None:
            # TODO
            return EnumData()
        data = cls_data.enums.get(name, _missing)
        if data is _missing:
            self.classes[cls_key]["enums"][name] = False
            data = EnumData()

        return data

    def get_enum_data(self, name: str) -> EnumData:
        data = self.data.enums.get(name, _missing)
        if data is _missing:
            self.enums[name] = False
            data = EnumData()
        return data

    def get_function_data(
        self,
        fn: dict,
        signature: str,
        cls_key: Optional[str] = None,
        cls_data: Optional[ClassData] = None,
        is_private: bool = False,
    ) -> FunctionData:
        name = fn["name"]
        if cls_data:
            data = cls_data.methods.get(name, _missing)
            report_base = self.classes[cls_key]["functions"]
        else:
            data = self.data.functions.get(name, _missing)
            report_base = self.functions

        report_base = report_base.setdefault(name, {"overloads": {}, "first": fn})
        missing = data is _missing
        report_base["missing"] = missing and not is_private

        if missing:
            data = FunctionData()
        else:
            overload = data.overloads.get(signature, _missing)
            missing = overload is _missing
            if not missing and overload:
                # merge overload information
                data = data.dict(exclude_unset=True)
                del data["overloads"]
                data.update(overload.dict(exclude_unset=True))
                data = FunctionData(**data)

        report_base["overloads"][signature] = is_private or not missing

        # TODO: doesn't belong here
        is_overloaded = len(report_base["overloads"]) > 1
        if is_overloaded:
            report_base["first"]["x_overloaded"] = True
        fn["x_overloaded"] = is_overloaded

        return data

    def get_cls_prop_data(
        self, name: str, cls_key: str, cls_data: ClassData
    ) -> PropData:
        data = cls_data.attributes.get(name, _missing)
        if data is _missing:
            self.classes[cls_key]["attributes"][name] = False
            data = PropData()

        return data

    def get_prop_data(self, name) -> PropData:
        data = self.data.attributes.get(name, _missing)
        if data is _missing:
            self.attributes[name] = False
            data = PropData()

        return data

    def report_missing(self, name: str, reporter: "MissingReporter"):
        """
            Generate a structure that can be copy/pasted into the generation
            data yaml and print it out if there's missing data
        """

        # note: sometimes we have strings from CppHeaderParser that aren't
        # strings, so we need to cast them to str so yaml doesn't complain

        data = self._process_missing(
            self.attributes, self.functions, self.enums, "functions"
        )

        all_cls_data = {}
        for cls_key, cls_data in self.classes.items():
            result = self._process_missing(
                cls_data["attributes"],
                cls_data["functions"],
                cls_data["enums"],
                "methods",
            )
            if result or cls_data["missing"]:
                # show this first
                r = {"shared_ptr": True}
                r.update(result)
                all_cls_data[str(cls_key)] = r
        if all_cls_data:
            data["classes"] = all_cls_data

        if data:
            reporter.add_report(name, data)

        return data

    def _process_missing(self, attrs, fns, enums, fn_key: str):

        data = {}

        # attributes
        if attrs:
            for y in attrs.keys():
                assert isinstance(y, str)

            data["attributes"] = {str(n): {} for n in attrs.keys()}

        # enums
        if enums:
            data["enums"] = {str(n): {} for n in enums.keys()}

        # functions
        fn_report = {}
        for fn, fndata in fns.items():
            fn = str(fn)
            overloads = fndata["overloads"]
            overloads_count = len(overloads)
            if overloads_count > 1:
                has_data = all(overloads.values())
            else:
                has_data = not fndata["missing"]

            if not has_data:
                d = {}
                if fn == "swap":
                    d = {"ignore": True}

                if overloads_count > 1:
                    fn_report[fn] = {
                        "overloads": {
                            k: dict(**d) for k, v in overloads.items() if not v
                        }
                    }

                    for k, v in fn_report[fn]["overloads"].items():
                        if "initializer_list" in k:
                            v["ignore"] = True
                else:
                    fn_report[fn] = d
        if fn_report:
            data[fn_key] = fn_report

        return data


class MissingReporter:
    def __init__(self):
        self.reports = {}

    def _merge(self, src, dst):
        for k, v in src.items():
            if isinstance(v, dict):
                if k not in dst:
                    dst[k] = v
                else:
                    self._merge(v, dst[k])
            else:
                dst[k] = v

    def add_report(self, name, data):
        if name in self.reports:
            self._merge(data, self.reports[name])
        else:
            self.reports[name] = data

    def as_yaml(self):
        for name, report in self.reports.items():
            yield name, (
                yaml.safe_dump(report, sort_keys=False)
                .replace(" {}", "")
                .replace("? ''\n          :", '"":')
                .replace("? ''\n      :", '"":')
            )



import os
from os.path import join
import posixpath
import shutil
from typing import Dict, List, Optional

from .download import download_maven
from .pyproject_configs import StaticLibConfig


class StaticLib:
    # implements pkgcfg

    def __init__(self, name: str, cfg: StaticLibConfig, setup):
        self.name = name
        self.cfg = cfg
        self.static_lib = True
        self.libinit_import = None
        # TODO
        self.depends = []

        self.platform = setup.platform

        self.root = None
        self.incdir = None
        self.libdir = None

    def set_root(self, root):
        self.root = root
        self.libdir = join(self.root, self.name, "lib")
        self.incdir = join(self.root, self.name, "include")

    def get_include_dirs(self) -> Optional[List[str]]:
        return [self.incdir]

    def get_library_dirs(self) -> Optional[List[str]]:
        return [self.libdir]

    def get_library_dirs_rel(self) -> Optional[List[str]]:
        pass

    def get_library_names(self) -> Optional[List[str]]:
        # don't do this except on Windows
        if self.platform.os != "windows":
            return

        return self._get_libnames(useext=False)

    def get_library_full_names(self) -> Optional[List[str]]:
        pass

    def get_extra_objects(self) -> Optional[List[str]]:
        if self.platform.os == "windows":
            return

        return [join(self.libdir, lib) for lib in self._get_libnames()]

    def get_type_casters(self, casters: Dict[str, str]) -> None:
        pass

    def _get_libnames(self, useext=True):
        mcfg = self.cfg.maven_lib_download
        if mcfg.libs is None:
            libs = [mcfg.artifact_id]
        else:
            libs = mcfg.libs
        ext = ""
        if useext:
            ext = self.platform.staticext
        return [f"{self.platform.libprefix}{lib}{ext}" for lib in libs]

    def on_build_dl(self, cache: str, libdir: str):

        dlcfg = self.cfg.maven_lib_download

        self.set_root(libdir)

        shutil.rmtree(self.libdir, ignore_errors=True)
        shutil.rmtree(self.incdir, ignore_errors=True)

        os.makedirs(self.libdir)

        download_maven(dlcfg, "headers", self.incdir, cache)

        extract_names = self._get_libnames()

        to = {
            posixpath.join(
                self.platform.os, self.platform.arch, "static", libname
            ): join(self.libdir, libname)
            for libname in extract_names
        }

        download_maven(
            dlcfg, f"{self.platform.os}{self.platform.arch}static", to, cache
        )



from distutils.util import get_platform as _get_platform
from dataclasses import dataclass, field
from typing import List, Tuple, Union
import re


# wpilib platforms at https://github.com/wpilibsuite/native-utils/blob/master/src/main/java/edu/wpi/first/nativeutils/WPINativeUtilsExtension.java
@dataclass
class WPILibMavenPlatform:
    arch: str
    os: str = "linux"
    libprefix: str = "lib"

    #: runtime linkage
    libext: str = ".so"

    #: compile time linkage
    linkext: str = None

    #: static linkage
    staticext: str = ".a"

    defines: List[str] = field(default_factory=list)

    def __post_init__(self):
        # linkext defaults to libext
        if self.linkext is None:
            self.linkext = self.libext
        self.defines = [f"{d} 1" for d in self.defines]


X86_64 = "x86-64"

# key is python platform, value is information about wpilib maven artifacts
_platforms = {
    # TODO: this isn't always true
    # -> __FRC_ROBORIO__ is injected by the WPILib built compiler
    "linux-armv7l": WPILibMavenPlatform("athena", defines=["__FRC_ROBORIO__"]),
    "linux-x86_64": WPILibMavenPlatform(X86_64),
    # TODO: linuxraspbian
    "win32": WPILibMavenPlatform("x86", "windows", "", ".dll", ".lib", ".lib"),
    "win-amd64": WPILibMavenPlatform(X86_64, "windows", "", ".dll", ".lib", ".lib"),
    "osx": WPILibMavenPlatform(X86_64, "osx", libext=".dylib"),
}


def get_platform() -> WPILibMavenPlatform:
    """
        Retrieve platform specific information
    """

    # TODO: _PYTHON_HOST_PLATFORM is used for cross builds,
    #       and is returned directly from get_platform. Might
    #       be useful to note for the future.

    pyplatform = _get_platform()

    # Check for 64 bit x86 macOS (version agnostic)
    if re.fullmatch(r"macosx-.*-x86_64", pyplatform):
        pyplatform = "osx"

    try:
        return _platforms[pyplatform]
    except KeyError:
        raise KeyError(f"platform {pyplatform} is not supported by robotpy-build!")


def get_platform_override_keys(platform: WPILibMavenPlatform):
    # Used in places where overrides exist
    return [
        f"arch_{platform.arch}",
        f"os_{platform.os}",
        f"platform_{platform.os}_{platform.arch}",
    ]



import atexit
import os
from os.path import exists, join
import posixpath
import shutil
import sys
import tempfile
import zipfile


from urllib.request import urlretrieve, urlcleanup


def _download(url):
    """
        Downloads a file to a temporary directory
    """

    print("Downloading", url)

    def _reporthook(count, blocksize, totalsize):
        percent = int(count * blocksize * 100 / totalsize)
        sys.stdout.write("\r%02d%%" % percent)
        sys.stdout.flush()

    filename, _ = urlretrieve(url, reporthook=_reporthook)
    atexit.register(urlcleanup)
    return filename


def download_and_extract_zip(url, to=None, cache=None):
    """
        Utility method intended to be useful for downloading/extracting
        third party source zipfiles

        :param to: is either a string or a dict of {src: dst}
    """

    if to is None:
        # generate temporary directory
        tod = tempfile.TemporaryDirectory()
        to = tod.name
        atexit.register(tod.cleanup)

    zip_fname = None
    if cache:
        os.makedirs(cache, exist_ok=True)
        cache_fname = join(cache, posixpath.basename(url))
        if not exists(cache_fname):
            zip_fname = _download(url)
            shutil.copy(zip_fname, cache_fname)
        zip_fname = cache_fname
    else:
        zip_fname = _download(url)

    with zipfile.ZipFile(zip_fname) as z:
        if isinstance(to, str):
            z.extractall(to)
            return to
        else:
            for src, dst in to.items():
                with z.open(src, "r") as zfp:
                    with open(dst, "wb") as fp:
                        shutil.copyfileobj(zfp, fp)


def download_maven(dlcfg, classifier, to, cache):
    # TODO: support development against locally installed things?
    repo_url = dlcfg.repo_url
    grp = dlcfg.group_id.replace(".", "/")
    art = dlcfg.artifact_id
    ver = dlcfg.version

    url = f"{repo_url}/{grp}/{art}/{ver}/{art}-{ver}-{classifier}.zip"
    return download_and_extract_zip(url, to, cache)



from setuptools.command.build_py import build_py


class BuildPy(build_py):

    wrappers = []

    def run(self):

        # files need to be generated before building can occur
        # -> otherwise they're not included in the bdist
        self.run_command("build_gen")

        # Add the generated files to the package data
        for package, _, _, filenames in self.data_files:
            for wrapper in self.wrappers:
                if wrapper.package_name == package:
                    filenames.extend(wrapper.generated_files)
                    break

        build_py.run(self)



from typing import Dict, List


def _merge(name, d, override_keys):
    override = d.pop("override", None)
    if override is not None:
        if not isinstance(override, dict):
            raise ValueError(
                f"{name}.override: got {type(override)}, must be a dictionary"
            )
        for k in override_keys:
            v = override.get(k, None)
            if v is not None:
                if not isinstance(v, dict):
                    raise ValueError(
                        f"{name}.override.{k}: got {type(v)}, must be a dictionary"
                    )
                d.update(v)

    for k, v in d.items():
        if isinstance(v, dict):
            _merge(f"{name}.{k}", v, override_keys)
        elif isinstance(v, list):
            for i in v:
                if isinstance(i, dict):
                    _merge(f"{name}.{k}[{i}]", i, override_keys)


def apply_overrides(d: Dict, override_keys: List[str]):
    """
        The idea is that any dictionary key can contain a key
        called 'override', which will be processed as follows:

        - override contents must be a dictionary
          - key is a potential 'override key'
          - value must be a dictionary
        - If a key matches an override key, then the contained
          dictionary will be shallow updated with the contents
          of the dictionary
        - This will recurse all structures and implement overrides
          wherever dictionaries are found

        Currently, if a list is overridden, the entire list is replaced. It may
        make sense to provide
    """
    _merge("", d, override_keys)


def _main():
    import argparse
    import toml

    parser = argparse.ArgumentParser()
    parser.add_argument("fname")
    parser.add_argument("key")
    args = parser.parse_args()

    with open(args.fname) as fp:
        d = toml.load(fp)

    apply_overrides(d, [args.key])
    print(toml.dumps(d))


if __name__ == "__main__":
    _main()



from distutils.core import Command
import os.path

from ..platforms import get_platform
from .util import get_install_root


class BuildDl(Command):

    command_name = "build_dl"
    description = "Downloads files"
    user_options = [
        ("build-base=", "b", "base directory for build library"),
        ("build-temp=", "t", "temporary build directory"),
        ("build-cache=", None, "build directory to cache downloaded objects"),
        ("src-unpack-to=", None, "build directory to unpack sources to"),
        ("lib-unpack-to=", None, "build directory to unpack static libs to"),
    ]
    wrappers = []
    static_libs = []

    def initialize_options(self):
        self.build_base = None
        self.build_cache = None
        self.build_temp = None
        self.src_unpack_to = None
        self.lib_unpack_to = None

    def finalize_options(self):
        self.set_undefined_options(
            "build", ("build_base", "build_base"), ("build_temp", "build_temp")
        )
        if self.build_cache is None:
            self.build_cache = os.path.join(self.build_base, "cache")
        if self.src_unpack_to is None:
            self.src_unpack_to = os.path.join(self.build_temp, "dlsrc")
        if self.lib_unpack_to is None:
            self.lib_unpack_to = os.path.join(self.build_temp, "dlstatic")

    def run(self):
        for lib in self.static_libs:
            lib.on_build_dl(self.build_cache, self.lib_unpack_to)
        for wrapper in self.wrappers:
            wrapper.on_build_dl(self.build_cache, self.src_unpack_to)

        # On OSX, fix library loader paths for embedded libraries
        # -> this happens here so that the libs are modified before build_py
        #    copies them. Extensions are fixed after build
        platform = get_platform()
        if platform.os == "osx":
            from ..relink_libs import relink_libs

            install_root = get_install_root(self)
            for wrapper in self.wrappers:
                relink_libs(install_root, wrapper, self.rpybuild_pkgcfg)
